"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Injector = void 0;
var tslib_1 = require("tslib");
var injectable_1 = require("./injectable");
var injection_error_1 = require("./injection-error");
var Injector = /** @class */ (function () {
    function Injector() {
        this.factories = new Map();
        this.resolved = new Map();
    }
    Injector_1 = Injector;
    /**
     * Create a child injector that inherits all factories and resolved values.
     * Providers may be added and dependencies overridden without effecting the parent injector.
     *
     * @param tokensToProvide An array of tokens the child injector should construct locally.
     */
    Injector.prototype.createChildInjector = function (tokensToProvide) {
        var childInjector = new Injector_1();
        childInjector.parentInjector = this;
        childInjector.factories = new Map();
        for (var _i = 0, tokensToProvide_1 = tokensToProvide; _i < tokensToProvide_1.length; _i++) {
            var token = tokensToProvide_1[_i];
            var injector = this;
            while (injector && !injector.factories.has(token)) {
                injector = injector.parentInjector;
            }
            if (!injector) {
                throw new injection_error_1.InjectionError("InjectionError: No provider for " + tokenName(token) + ".", token);
            }
            var factory = injector.factories.get(token);
            childInjector.factories.set(token, factory);
        }
        childInjector.factories.set(Injector_1, { factory: function () { return childInjector; }, dependencies: [] });
        childInjector.resolved.set(Injector_1, childInjector);
        return childInjector;
    };
    /** Override a dependency token for unit tests */
    Injector.prototype.overrideDependency = function (token, value) {
        this.factories.set(token, { factory: function () { return value; }, dependencies: [] });
        this.resolved.set(token, value);
    };
    Injector.prototype.get = function (token, notFoundValue) {
        if (this.resolved.has(token)) {
            return this.resolved.get(token);
        }
        if (token === Injector_1) {
            return this;
        }
        var injector = this;
        do {
            if (injector.resolved.has(token)) {
                var resolvedValue = injector.resolved.get(token);
                this.resolved.set(token, resolvedValue);
                return resolvedValue;
            }
            if (injector.factories.has(token)) {
                var factory = injector.factories.get(token);
                var dependencies = factory.dependencies.map(function (dep) { return injector.get(dep); });
                var resolvedValue = factory.factory.apply(factory, dependencies);
                injector.resolved.set(token, resolvedValue);
                this.resolved.set(token, resolvedValue);
                return resolvedValue;
            }
        } while (injector = injector.parentInjector);
        if (arguments.length > 1) {
            return notFoundValue;
        }
        throw new injection_error_1.InjectionError("InjectionError: No provider for " + tokenName(token) + ".", token);
    };
    Injector.isValidProvider = function (input) {
        if (typeof input === 'function' && input.prototype && input.prototype.constructor === input) {
            return true;
        }
        else if (typeof input === 'object' && input && input.provide) {
            return true;
        }
        else {
            return false;
        }
    };
    /**
     * Resolve the all dependencies of the passed providers as dependency tree
     * and return an injector that creates and returns instances of the dependencies.
     *
     * All classes that are not explicitly provided are resolved to the class itself,
     * missing InjectionToken providers throw an exception.
     *
     * @example
     *     const injector = Injector.autoResolveAndCreate([App]);
     */
    Injector.autoResolveAndCreate = function (providers) {
        validateProviders(providers);
        var injector = new Injector_1();
        var factories = injector.factories = factoriesFromProviders(providers);
        var dependencies = Array.from(factories.keys());
        throwOnCircularDependencies(dependencies, factories, []);
        addMissingClassDependencies(dependencies, factories);
        throwOnCircularDependencies(dependencies, factories, []);
        throwOnMissingDependencies(dependencies, factories);
        return injector;
    };
    /**
     * Resolve the passed providers as dependency tree and return an injector
     * that creates and returns instances of the dependencies.
     *
     * Providers for all nested dependencies need to be in the input array.
     *
     * @example
     *     const injector = Injector.resolveAndCreate([App, Dependency1, Dependency2]);
     */
    Injector.resolveAndCreate = function (providers) {
        validateProviders(providers);
        var injector = new Injector_1();
        var factories = injector.factories = factoriesFromProviders(providers);
        var dependencies = Array.from(factories.keys());
        throwOnCircularDependencies(dependencies, factories, []);
        throwOnMissingDependencies(dependencies, factories);
        return injector;
    };
    var Injector_1;
    Injector = Injector_1 = tslib_1.__decorate([
        injectable_1.Injectable
    ], Injector);
    return Injector;
}());
exports.Injector = Injector;
function tokenName(token) {
    return token && token.name || String(token);
}
function getDiTokens(classConstructor) {
    var metadata = Reflect.getMetadata('di-tokens', classConstructor);
    if (metadata) {
        return metadata;
    }
    else if (!classConstructor.length) {
        return [];
    }
    throw new injection_error_1.InjectionError('Class with dependencies is provided but not decorated with @Injectable or @Inject', classConstructor);
}
/** Normalize providers to a `{ factory, dependencies }` hash. */
function factoriesFromProviders(providers) {
    var factories = new Map();
    var _loop_1 = function (provider) {
        var token = provider.provide;
        if (!token && provider.prototype && provider.prototype.constructor === provider) {
            // Provider is a class
            var dependencies = getDiTokens(provider);
            factories.set(provider, {
                factory: function () {
                    var args = [];
                    for (var _i = 0; _i < arguments.length; _i++) {
                        args[_i] = arguments[_i];
                    }
                    return new (provider.bind.apply(provider, tslib_1.__spreadArray([void 0], args)))();
                },
                dependencies: dependencies
            });
        }
        else if ('useValue' in provider) {
            // Provider is a { provide: ClassName, useValue: ... } object
            var valueToUse_1 = provider.useValue;
            factories.set(token, {
                factory: function () { return valueToUse_1; },
                dependencies: []
            });
        }
        else if ('useFactory' in provider) {
            // Provider is a { provide: ClassName, useFactory: ... } object
            factories.set(token, {
                factory: provider.useFactory,
                dependencies: provider.dependencies || []
            });
        }
        else if ('useClass' in provider) {
            // Provider is a { provide: ClassName, useClass: ... } object
            var classToUse_1 = provider.useClass;
            factories.set(token, {
                factory: function () {
                    var args = [];
                    for (var _i = 0; _i < arguments.length; _i++) {
                        args[_i] = arguments[_i];
                    }
                    return new (classToUse_1.bind.apply(classToUse_1, tslib_1.__spreadArray([void 0], args)))();
                },
                dependencies: provider.dependencies || getDiTokens(classToUse_1)
            });
        }
    };
    for (var _i = 0, _a = providers; _i < _a.length; _i++) {
        var provider = _a[_i];
        _loop_1(provider);
    }
    return factories;
}
/** Throw if a provider is not a class or a `{ provide: ... }` hash */
function validateProviders(providers) {
    for (var _i = 0, _a = providers; _i < _a.length; _i++) {
        var p = _a[_i];
        if (!Injector.isValidProvider(p)) {
            throw new Error("Invalid provider: " + p);
        }
    }
}
function throwOnCircularDependencies(dependencies, factories, dependents) {
    for (var _i = 0, dependencies_1 = dependencies; _i < dependencies_1.length; _i++) {
        var dependency = dependencies_1[_i];
        if (dependents.indexOf(dependency) >= 0) {
            throw new injection_error_1.InjectionError("Circular dependency: \"" + tokenName(dependency) + "\" is required by one of its dependencies.", dependency);
        }
        var factory = factories.get(dependency);
        if (factory) {
            throwOnCircularDependencies(factory.dependencies, factories, tslib_1.__spreadArray(tslib_1.__spreadArray([], dependents), [dependency]));
        }
    }
}
function throwOnMissingDependencies(dependencies, factories) {
    for (var _i = 0, dependencies_2 = dependencies; _i < dependencies_2.length; _i++) {
        var dependency = dependencies_2[_i];
        if (!factories.has(dependency)) {
            throw new injection_error_1.InjectionError("No provider for \"" + tokenName(dependency) + "\".", dependency);
        }
        var factory = factories.get(dependency);
        throwOnMissingDependencies(factory.dependencies, factories);
    }
}
function addMissingClassDependencies(dependencies, factories) {
    var _loop_2 = function (dependency) {
        var factory = factories.get(dependency);
        if (!factory) {
            if (typeof dependency === 'function' && dependency.prototype && dependency.prototype.constructor === dependency) {
                // dependency is a class constructor, add a factory for it
                var classDependencies = getDiTokens(dependency);
                factories.set(dependency, factory = {
                    factory: function () {
                        var args = [];
                        for (var _i = 0; _i < arguments.length; _i++) {
                            args[_i] = arguments[_i];
                        }
                        return new (dependency.bind.apply(dependency, tslib_1.__spreadArray([void 0], args)))();
                    },
                    dependencies: classDependencies
                });
            }
            else {
                // dependency is an InjectionToken
                throw new injection_error_1.InjectionError("No provider for \"" + tokenName(dependency) + "\".", dependency);
            }
        }
        addMissingClassDependencies(factory.dependencies, factories);
    };
    for (var _i = 0, dependencies_3 = dependencies; _i < dependencies_3.length; _i++) {
        var dependency = dependencies_3[_i];
        _loop_2(dependency);
    }
}
//# sourceMappingURL=injector.js.map